"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const rxjs_1 = require("rxjs");
const operators_1 = require("rxjs/operators");
const rabbitmq_channel_cancelled_error_1 = require("../errors/rabbitmq-channel-cancelled.error");
const rabbitmq_channel_closed_error_1 = require("../errors/rabbitmq-channel-closed.error");
const rabbitmq_channel_error_1 = require("../errors/rabbitmq-channel.error");
const rabbitmq_connection_closed_error_1 = require("../errors/rabbitmq-connection-closed.error");
const rabbitmq_connection_error_1 = require("../errors/rabbitmq-connection.error");
const rabbitmq_reconnect_error_1 = require("../errors/rabbitmq-reconnect.error");
const rabbit_mq_connection_factory_1 = require("../factories/rabbit-mq-connection-factory");
const DEFAULT_RECONNECT_TIMEOUT_MILLIS = 1000;
const DEFAULT_RECONNECT_ATTEMPTS = -1;
const DEFAULT_PREFETCH = 100;
class Consumer {
    constructor() { }
    fillEmptyConfigsWithDefaults(rawConfigs) {
        let filledConfigs = Object.assign({}, rawConfigs);
        if (!filledConfigs.queue || !filledConfigs.queue.name)
            throw new Error('Name of the queue has to be provided');
        filledConfigs.queue.options = filledConfigs.queue.options || { durable: true };
        if (typeof filledConfigs.queue.options.durable === 'undefined')
            filledConfigs.queue.options.durable = true;
        filledConfigs.exchange = filledConfigs.exchange || {};
        filledConfigs.exchange.name = filledConfigs.exchange.name || `exchange_${filledConfigs.queue.name}`;
        filledConfigs.exchange.options = filledConfigs.exchange.options || { durable: true };
        if (typeof filledConfigs.exchange.options.durable === 'undefined')
            filledConfigs.exchange.options.durable = true;
        filledConfigs.exchange.type = filledConfigs.exchange.type || 'direct';
        if (filledConfigs.exchange.type === 'topic' && !filledConfigs.queue.bindingPattern)
            throw new Error('In case of topic exchanges, routing pattern has to be provided for the queue');
        if (filledConfigs.exchange.type === 'direct')
            filledConfigs.queue.bindingPattern = filledConfigs.queue.bindingPattern || `${filledConfigs.queue.name}_rk`;
        else if (filledConfigs.exchange.type === 'fanout')
            filledConfigs.queue.bindingPattern = '';
        filledConfigs.autoAck = typeof filledConfigs.autoAck === 'undefined' ? false : filledConfigs.autoAck;
        if (filledConfigs.prefetch < 1)
            throw new Error('Prefetch count should be at least 1');
        filledConfigs.prefetch = filledConfigs.prefetch || DEFAULT_PREFETCH;
        if (filledConfigs.reconnectAttempts < 0 && filledConfigs.reconnectAttempts !== -1)
            throw new Error('Reconnect attempts count should be at least 0');
        filledConfigs.reconnectAttempts = filledConfigs.reconnectAttempts || DEFAULT_RECONNECT_ATTEMPTS;
        if (filledConfigs.reconnectTimeoutMillis < 0)
            throw new Error('Reconnect timeout should be at least 0 ms');
        filledConfigs.reconnectTimeoutMillis = filledConfigs.reconnectTimeoutMillis || DEFAULT_RECONNECT_TIMEOUT_MILLIS;
        if (typeof filledConfigs.noDeadLetterQueue !== 'undefined' && filledConfigs.noDeadLetterQueue === true) {
            delete filledConfigs.deadLetterQueue;
            delete filledConfigs.noDeadLetterQueue;
        }
        else {
            filledConfigs.deadLetterQueue = filledConfigs.deadLetterQueue || {};
            filledConfigs.deadLetterQueue.queue = filledConfigs.deadLetterQueue.queue || {};
            filledConfigs.deadLetterQueue.queue.name = filledConfigs.deadLetterQueue.queue.name || `dlq_${filledConfigs.queue.name}`;
            filledConfigs.deadLetterQueue.queue.options = filledConfigs.deadLetterQueue.queue.options || { durable: true };
            filledConfigs.deadLetterQueue.exchange = filledConfigs.deadLetterQueue.exchange || {};
            filledConfigs.deadLetterQueue.exchange.name = filledConfigs.deadLetterQueue.exchange.name || `exchange_${filledConfigs.deadLetterQueue.queue.name}`;
            filledConfigs.deadLetterQueue.exchange.type = filledConfigs.deadLetterQueue.exchange.type || 'direct';
            filledConfigs.deadLetterQueue.exchange.options = filledConfigs.deadLetterQueue.exchange.options || { durable: true };
            if (filledConfigs.deadLetterQueue.exchange.type === 'topic' && !filledConfigs.deadLetterQueue.queue.bindingPattern)
                throw new Error('In case of topic exchanges, routing pattern has to be provided for the queue');
            if (filledConfigs.deadLetterQueue.exchange.type === 'direct')
                filledConfigs.deadLetterQueue.queue.bindingPattern = filledConfigs.deadLetterQueue.queue.bindingPattern || `${filledConfigs.deadLetterQueue.queue.name}_rk`;
            else if (filledConfigs.deadLetterQueue.exchange.type === 'fanout')
                filledConfigs.deadLetterQueue.queue.bindingPattern = '';
        }
        return filledConfigs;
    }
    defaultSetup(connection) {
        return __awaiter(this, void 0, void 0, function* () {
            const channel = yield connection.createChannel();
            if (this.configs.deadLetterQueue) {
                yield channel.assertExchange(this.configs.deadLetterQueue.exchange.name, this.configs.deadLetterQueue.exchange.type, this.configs.deadLetterQueue.exchange.options);
                const dlqMetadata = yield channel.assertQueue(this.configs.deadLetterQueue.queue.name, this.configs.deadLetterQueue.queue.options);
                yield channel.bindQueue(dlqMetadata.queue, this.configs.deadLetterQueue.exchange.name, this.configs.deadLetterQueue.queue.bindingPattern);
            }
            yield channel.assertExchange(this.configs.exchange.name, this.configs.exchange.type, this.configs.exchange.options);
            const queueMetadata = yield channel.assertQueue(this.configs.queue.name, Object.assign({}, this.configs.queue.options, { deadLetterExchange: this.configs.deadLetterQueue ? this.configs.deadLetterQueue.exchange.name : undefined, deadLetterRoutingKey: this.configs.deadLetterQueue ? this.configs.deadLetterQueue.queue.bindingPattern : undefined }));
            yield channel.bindQueue(queueMetadata.queue, this.configs.exchange.name, this.configs.queue.bindingPattern, this.configs.queue.options.arguments);
            yield channel.prefetch(this.configs.prefetch);
            return channel;
        });
    }
    setConfigs(configs) {
        this.rawConfigs = configs;
        this.configs = this.fillEmptyConfigsWithDefaults(this.rawConfigs);
    }
    setCustomSetupFunction(setupFunction) {
        this.customSetupFunction = setupFunction;
        this.rawConfigs = null;
        this.configs = null;
    }
    init(connection) {
        return __awaiter(this, void 0, void 0, function* () {
            this.connection = connection;
            const amqpConnection = this.connection.getAmqpConnection();
            let channel;
            let prefetch;
            let autoAck;
            let queue;
            if (this.customSetupFunction) {
                const functionResult = yield this.customSetupFunction(amqpConnection);
                channel = functionResult.channel;
                prefetch = functionResult.prefetch;
                autoAck = functionResult.autoAck;
                queue = functionResult.queue;
                this.channel = channel;
                this.subject = new rxjs_1.ReplaySubject(prefetch || DEFAULT_PREFETCH);
            }
            else {
                this.channel = yield this.defaultSetup(amqpConnection);
                this.subject = new rxjs_1.ReplaySubject(this.configs.prefetch);
                queue = this.configs.queue.name;
            }
            yield this.channel.consume(queue, (message) => {
                if (message === null)
                    return void this.subject.error(new rabbitmq_channel_cancelled_error_1.RabbitMqChannelCancelledError('The channel was cancelled by the server'));
                else
                    this.subject.next(message);
            }, { noAck: autoAck });
            amqpConnection.on('error', (err) => this.subject.error(new rabbitmq_connection_error_1.RabbitMqConnectionError(err.message)));
            amqpConnection.on('close', () => this.subject.error(new rabbitmq_connection_closed_error_1.RabbitMqConnectionClosedError('The connection was closed by the server')));
            this.channel.on('error', (err) => this.subject.error(new rabbitmq_channel_error_1.RabbitMqChannelError(err.message)));
            this.channel.on('close', () => this.subject.error(new rabbitmq_channel_closed_error_1.RabbitMqChannelClosedError('The channel was closed by the server')));
        });
    }
    reconnect() {
        const connectionFactory = new rabbit_mq_connection_factory_1.RabbitMqConnectionFactory();
        if (this.connection.getUri())
            connectionFactory.setUri(this.connection.getUri());
        else
            connectionFactory.setOptions(this.connection.getOptions());
        return new rxjs_1.Observable((subscriber) => {
            connectionFactory
                .newConnection()
                .then((connection) => this.init(connection))
                .then(() => subscriber.complete())
                .catch(() => { });
        }).pipe(operators_1.timeout(this.configs.reconnectTimeoutMillis || DEFAULT_RECONNECT_TIMEOUT_MILLIS), operators_1.retry(this.configs.reconnectAttempts ? this.configs.reconnectAttempts - 1 : DEFAULT_RECONNECT_ATTEMPTS)).pipe(operators_1.catchError(() => {
            throw new rabbitmq_reconnect_error_1.RabbitMqReconnectError('Failed to reconnect');
        }));
    }
    closeChannel() {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.channel.close();
        });
    }
    getActiveChannel() {
        return this.channel;
    }
    getActiveConnection() {
        return this.connection.getAmqpConnection();
    }
    startConsuming() {
        return this.subject;
    }
    getActiveConfigs() {
        return this.configs;
    }
    commitMessage(amqpMessage) {
        this.channel.ack(amqpMessage);
    }
    rejectMessage(message, allUpToCurrent = false, requeue = false) {
        this.channel.nack(message, allUpToCurrent, requeue);
    }
}
exports.Consumer = Consumer;
//# sourceMappingURL=consumer.js.map