"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const BehaviorSubject_1 = require("rxjs/internal/BehaviorSubject");
const Observable_1 = require("rxjs/internal/Observable");
const operators_1 = require("rxjs/operators");
const rabbitmq_channel_closed_error_1 = require("../errors/rabbitmq-channel-closed.error");
const rabbitmq_channel_error_1 = require("../errors/rabbitmq-channel.error");
const rabbitmq_connection_closed_error_1 = require("../errors/rabbitmq-connection-closed.error");
const rabbitmq_connection_error_1 = require("../errors/rabbitmq-connection.error");
const rabbitmq_publisher_confirmation_error_1 = require("../errors/rabbitmq-publisher-confirmation.error");
const rabbitmq_reconnect_error_1 = require("../errors/rabbitmq-reconnect.error");
const rabbit_mq_connection_factory_1 = require("../factories/rabbit-mq-connection-factory");
const DEFAULT_RECONNECT_TIMEOUT_MILLIS = 1000;
const DEFAULT_RECONNECT_ATTEMPTS = -1;
class Publisher {
    constructor() { }
    fillEmptyConfigsWithDefaults(rawConfigs) {
        let filledConfigs = Object.assign({}, rawConfigs);
        if (!filledConfigs.exchange || !filledConfigs.exchange.name)
            throw new Error('Name of the exchange has to be provided');
        filledConfigs.exchange.options = filledConfigs.exchange.options || { durable: true };
        if (typeof filledConfigs.exchange.options.durable === 'undefined')
            filledConfigs.exchange.options.durable = true;
        filledConfigs.exchange.type = filledConfigs.exchange.type || 'direct';
        filledConfigs.publisherConfirms = typeof filledConfigs.publisherConfirms === 'undefined' ? true : filledConfigs.publisherConfirms;
        if (filledConfigs.reconnectAttempts < 0 && filledConfigs.reconnectAttempts !== -1)
            throw new Error('Reconnect attempts count should be at least 0');
        filledConfigs.reconnectAttempts = filledConfigs.reconnectAttempts || DEFAULT_RECONNECT_ATTEMPTS;
        if (filledConfigs.reconnectTimeoutMillis < 0)
            throw new Error('Reconnect timeout should be at least 0 ms');
        filledConfigs.reconnectTimeoutMillis = filledConfigs.reconnectTimeoutMillis || DEFAULT_RECONNECT_TIMEOUT_MILLIS;
        return filledConfigs;
    }
    defaultSetup(connection) {
        return __awaiter(this, void 0, void 0, function* () {
            const channel = this.configs.publisherConfirms ? yield connection.createConfirmChannel() : yield connection.createChannel();
            yield channel.assertExchange(this.configs.exchange.name, this.configs.exchange.type, this.configs.exchange.options);
            return channel;
        });
    }
    setConfigs(configs) {
        this.rawConfigs = configs;
        this.configs = this.fillEmptyConfigsWithDefaults(this.rawConfigs);
    }
    setCustomSetupFunction(setupFunction) {
        this.customSetupFunction = setupFunction;
        this.rawConfigs = null;
        this.configs = null;
    }
    init(connection) {
        return __awaiter(this, void 0, void 0, function* () {
            this.connection = connection;
            const amqpConnection = this.connection.getAmqpConnection();
            if (this.customSetupFunction) {
                const { channel, exchange, publisherConfirms } = yield this.customSetupFunction(amqpConnection);
                this.channel = channel;
                this.configs = {
                    exchange: {
                        name: exchange,
                    },
                    publisherConfirms,
                };
            }
            else
                this.channel = yield this.defaultSetup(amqpConnection);
            this.subject = new BehaviorSubject_1.BehaviorSubject('Publisher initialized');
            amqpConnection.on('error', (err) => this.subject.error(new rabbitmq_connection_error_1.RabbitMqConnectionError(err.message)));
            amqpConnection.on('close', () => this.subject.error(new rabbitmq_connection_closed_error_1.RabbitMqConnectionClosedError('AMQP server closed connection')));
            this.channel.on('error', (err) => this.subject.error(new rabbitmq_channel_error_1.RabbitMqChannelError(err.message)));
            this.channel.on('close', () => this.subject.error(new rabbitmq_channel_closed_error_1.RabbitMqChannelClosedError('AMQP server closed channel')));
        });
    }
    reconnect() {
        const connectionFactory = new rabbit_mq_connection_factory_1.RabbitMqConnectionFactory();
        if (this.connection.getUri())
            connectionFactory.setUri(this.connection.getUri());
        else
            connectionFactory.setOptions(this.connection.getOptions());
        return new Observable_1.Observable((subscriber) => {
            connectionFactory
                .newConnection()
                .then((connection) => this.init(connection))
                .then(() => subscriber.complete())
                .catch(() => { });
        }).pipe(operators_1.timeout(this.configs.reconnectTimeoutMillis || DEFAULT_RECONNECT_TIMEOUT_MILLIS), operators_1.retry(this.configs.reconnectAttempts ? this.configs.reconnectAttempts - 1 : DEFAULT_RECONNECT_ATTEMPTS)).pipe(operators_1.catchError(() => {
            throw new rabbitmq_reconnect_error_1.RabbitMqReconnectError('Failed to reconnect');
        }));
    }
    closeChannel() {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.channel.close();
        });
    }
    getActiveChannel() {
        return this.channel;
    }
    getActiveConnection() {
        return this.connection.getAmqpConnection();
    }
    getActiveConfigs() {
        return this.configs;
    }
    actionsStream() {
        return this.subject;
    }
    publishMessage(message, routingKey, options) {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.configs.publisherConfirms)
                return new Promise((resolve, reject) => {
                    this.channel.publish(this.configs.exchange.name, routingKey || '', message, options || {}, (err) => {
                        if (err)
                            return reject(new rabbitmq_publisher_confirmation_error_1.RabbitMqPublisherConfirmationError(err.message));
                        this.subject.next(`A message was sent with routing key "${routingKey}"`);
                        resolve();
                    });
                });
            this.channel.publish(this.configs.exchange.name, routingKey || '', message, options || {});
            this.subject.next(`A message was sent with routing key "${routingKey}"`);
        });
    }
}
exports.Publisher = Publisher;
//# sourceMappingURL=publisher.js.map